@@ -24,8 +24,13 @@
 ;;; Code:
 
 (define-module (json parser)
+  #:use-module (ice-9 rdelim)
+  #:use-module (ice-9 streams)
+  #:use-module (rnrs io ports)
   #:export (json->scm
-            json-string->scm))
+            json-string->scm
+            json-seq->scm
+            json-seq-string->scm))
 
 ;;
 ;; Miscellaneuos helpers
@@ -44,6 +49,9 @@
     ((#\sp #\ht #\lf #\cr) #t)
     (else #f)))
 
+(define (control-char? ch)
+  (<= (char->integer ch) #x1F))
+
 (define (skip-whitespaces port)
   (let ((ch (peek-char port)))
     (cond
@@ -70,6 +78,11 @@
      ;; Unexpected EOF.
      ((eof-object? ch) (json-exception port)))))
 
+(define (skip-record-separators port)
+  (when (eqv? #\rs (peek-char port))
+    (read-char port)
+    (skip-record-separators port)))
+
 ;;
 ;; Number parsing helpers
 ;;
@@ -134,7 +147,11 @@
     (cond
      ((or (eqv? ch #\e) (eqv? ch #\E))
       (read-char port)
-      (expt 10 (* (read-sign port) (read-digits port))))
+      (let ((sign (read-sign port))
+            (digits (read-digits port)))
+        (if (<= digits 1000) ;; Some maximum exponent.
+            (expt 10 (* sign digits))
+            (json-exception port))))
      (else 1))))
 
 (define (read-fraction port)
@@ -224,28 +241,30 @@
 
 (define (json-read-array port null)
   (expect-delimiter port #\[)
-  (let loop ((values '()) (added #t))
-    (skip-whitespaces port)
-    (let ((ch (peek-char port)))
-      (cond
-       ;; Unexpected EOF.
-       ((eof-object? ch) (json-exception port))
-       ;; Handle comma (make sure we added an element).
-       ((eqv? ch #\,)
-        (read-char port)
+  (skip-whitespaces port)
+  (cond
+   ;; Special case when array is empty.
+   ((eqv? (peek-char port) #\])
+    (read-char port)
+    #())
+   (else
+    ;; Read first element in array.
+    (let loop ((values (list (json-read port null))))
+      (skip-whitespaces port)
+      (let ((ch (peek-char port)))
         (cond
-         (added (loop values #f))
-         (else (json-exception port))))
-       ;; End of array (make sure we added an element).
-       ((eqv? ch #\])
-        (read-char port)
-        (cond
-         (added (list->vector (reverse! values)))
-         (else (json-exception port))))
-       ;; This can be any JSON object.
-       (else
-        (let ((value (json-read port null)))
-          (loop (cons value values) #t)))))))
+         ;; Unexpected EOF.
+         ((eof-object? ch) (json-exception port))
+         ;; Handle comma (if there's a comma there should be another element).
+         ((eqv? ch #\,)
+          (read-char port)
+          (loop (cons (json-read port null) values)))
+         ;; End of array.
+         ((eqv? ch #\])
+          (read-char port)
+          (list->vector (reverse! values)))
+         ;; Anything else other than comma and end of array is wrong.
+         (else (json-exception port))))))))
 
 ;;
 ;; String parsing helpers
@@ -325,6 +344,8 @@
     (cond
      ;; Unexpected EOF.
      ((eof-object? ch) (json-exception port))
+     ;; Unescaped control characters are not allowed.
+     ((control-char? ch) (json-exception port))
      ;; End of string.
      ((eqv? ch #\") (reverse-list->string chars))
      ;; Escaped characters.
@@ -372,19 +393,24 @@
 ;;
 
 (define* (json->scm #:optional (port (current-input-port))
-                    #:key (null 'null))
+                    #:key (null 'null) (concatenated #f))
   "Parse a JSON document into native. Takes one optional argument,
 @var{port}, which defaults to the current input port from where the JSON
-document is read. It also takes a keyword argument: @{null}: value for JSON's
-null, it defaults to the 'null symbol."
-  (let ((value (json-read port null)))
+document is read. It also takes a couple of keyword arguments: @{null}: value
+for JSON's null, it defaults to the 'null symbol and @{concatenated} which can
+be used to tell the parser that more JSON documents might come after a properly
+parsed document."
+  (let loop ((value (json-read port null)))
     ;; Skip any trailing whitespaces.
     (skip-whitespaces port)
     (cond
      ;; If we reach the end the parsing succeeded.
      ((eof-object? (peek-char port)) value)
-     ;; If there's anything else other than the end, parser fails.
-     (else (json-exception port)))))
+     ;; If there's anything else other than the end, check if user wants to keep
+     ;; parsing concatenated valid JSON documents, otherwise parser fails.
+     (else
+      (cond (concatenated value)
+            (else (json-exception port)))))))
 
 (define* (json-string->scm str #:key (null 'null))
   "Parse a JSON document into native. Takes a string argument,
@@ -392,4 +418,60 @@ null, it defaults to the 'null symbol."
 @{null}: value for JSON's null, it defaults to the 'null symbol "
   (call-with-input-string str (lambda (p) (json->scm p #:null null))))
 
+(define* (json-seq->scm #:optional (port (current-input-port))
+                        #:key (null 'null) (handle-truncate 'skip)
+                        (truncated-object 'truncated))
+  "Lazy parse a JSON text sequence from the port @var{port}.
+This procedure returns a stream of parsed documents. The optional argument
+@var{port} defines the port to read from and defaults to the current input
+port. It also takes a few keyword arguments: @{null}: value for JSON's null
+(defaults to the 'null symbol), @{handle-truncate}: defines how to handle data
+loss, @{truncated-object}: used to replace unparsable objects. Allowed values
+for @{handle-truncate} argument are 'throw (throw an exception), 'stop (stop
+parsing and end the stream), 'skip (default, skip corrupted fragment and
+return the next entry), 'replace (skip corrupted fragment and return
+@{truncated-object} instead)."
+  (letrec ((handle-truncation
+            (case handle-truncate
+              ((throw) (json-exception port))
+              ((stop) (const (eof-object)))
+              ((skip)
+               (lambda (port)
+                 (read-delimited "\x1e" port 'peek)
+                 (read-entry port)))
+              ((replace)
+               (lambda (port)
+                 (read-delimited "\x1e" port 'peek)
+                 truncated-object))))
+           (read-entry
+            (lambda (port)
+              (let ((ch (read-char port)))
+                (cond
+                 ((eof-object? ch) ch)
+                 ((not (eqv? ch #\rs))
+                  (handle-truncation port))
+                 (else
+                  (skip-record-separators port)
+                  (catch 'json-invalid
+                    (lambda ()
+                      (let ((next (json-read port null)))
+                        (if (eqv? #\lf (peek-char port))
+                            (begin
+                              (read-char port)
+                              next)
+                            (handle-truncation port))))
+                    (lambda (_ port)
+                      (handle-truncation port)))))))))
+    (port->stream port read-entry)))
+
+(define* (json-seq-string->scm str #:key (null 'null) (handle-truncate 'skip)
+                               (truncated-object 'truncated))
+  "Lazy parse a JSON text sequence from the string @var{str}.
+This procedure returns a stream of parsed documents and also takes the same
+keyword arguments as @code{json-seq->scm}."
+  (call-with-input-string str
+    (lambda (p)
+      (json-seq->scm p #:null null #:handle-truncate handle-truncate
+                     #:truncated-object truncated-object))))
+
 ;;; (json parser) ends here
