@@ -1,6 +1,6 @@
 ;;; (tests test-parser) --- Guile JSON implementation.
 
-;; Copyright (C) 2018-2020 Aleix Conchillo Flaque <aconchillo@gmail.com>
+;; Copyright (C) 2018-2021 Aleix Conchillo Flaque <aconchillo@gmail.com>
 ;;
 ;; This file is part of guile-json.
 ;;
@@ -24,8 +24,9 @@
 ;;; Code:
 
 (define-module (tests test-parser)
-  #:use-module (srfi srfi-64)
+  #:use-module (ice-9 streams)
   #:use-module (json)
+  #:use-module (srfi srfi-64)
   #:use-module (tests runner))
 
 (test-runner-factory json:test-runner)
@@ -111,6 +112,40 @@
 (test-error #t (json-string->scm "[1,2,3] extra"))
 (test-error #t (json-string->scm "{} extra"))
 
-(exit (if (test-end "test-parser") 0 1))
+;; Concatenated documents
+(call-with-input-string "{}12345 \"concatenated\"    {\"foo\":\"bar\"}"
+  (lambda (p)
+    (test-equal '() (json->scm p #:concatenated #t))
+    (test-equal '12345 (json->scm p #:concatenated #t))
+    (test-equal "concatenated" (json->scm p #:concatenated #t))
+    (test-equal '(("foo" . "bar")) (json->scm p #:concatenated #t))
+    (test-error #t (json->scm p #:concatenated #t))))
+
+(call-with-input-string "{} \"foo\" bar"
+  (lambda (p)
+    (test-equal '() (json->scm p #:concatenated #t))
+    (test-equal "foo" (json->scm p #:concatenated #t))
+    (test-error #t (json->scm p #:concatenated #t))))
+
+(call-with-input-string "{"
+  (lambda (p)
+    (test-error #t (json->scm p #:concatenated #t))))
+
+;; Sequences
+(define stream (json-seq-string->scm "\x1e1\n\x1e{}\n\x1e[null,true,false]\n\x1e{data loss"
+                                     #:handle-truncate 'replace))
+(test-equal 1 (stream-car stream))
+(test-equal '() (stream-car (stream-cdr stream)))
+(test-equal #(null #t #f) (stream-car (stream-cdr (stream-cdr stream))))
+(test-equal 'truncated (stream-car (stream-cdr (stream-cdr (stream-cdr stream)))))
+(test-assert (stream-null? (stream-cdr (stream-cdr (stream-cdr (stream-cdr stream))))))
+
+(test-assert (stream-null? (json-seq-string->scm "\x1enull\x1enull\x1enull" #:handle-truncate 'skip)))
+
+(test-equal "\x1e" (stream-car (json-seq-string->scm "\x1e\"\\u001e\"\n")))
+
+(let ((fail-count (test-runner-fail-count (test-runner-current))))
+  (test-end "test-parser")
+  (exit (zero? fail-count)))
 
 ;;; (tests test-parser) ends here
